/*********************************************************************************/
/* HIGA4DOP.cpp                                                                  */
/*                                                                               */
/* This is an implementation of genetic algorithms with immigrants schemes       */
/* for dynamic optimisation problems with the XOR generator for the paper:       */
/*                                                                               */
/*   S. Yang and R. Tinos. A hybrid immigrants scheme for genetic algorithms     */
/*   in dynamic environments. Int Journal of Automation and Computing,           */
/*   4(3): 243-254, July 2007. Springer (DOI: 10.1007/s11633-007-0243-9).        */
/*                                                                               */
/* Compile:                                                                      */
/*   make -f Makefile                                                            */
/*                                                                               */
/* Run:                                                                          */
/*   ./HIGA4DOP gaMode functNo chgMode chgDeg chgSpeed immRatio pImmMut pBaseMut */
/*                                                                               */
/* Written by:                                                                   */
/*   Shengxiang Yang, University of Leicester, UK, April 2007                    */
/*                                                                               */
/* If any query, please email Shengxiang Yang at s.yang@mcs.le.ac.uk             */
/*                                                                               */
/*********************************************************************************/

#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <string>
#include <fstream>

using namespace std;

/****************************************************************************
 *		  	display information of a pop		            *
 ****************************************************************************/
void display(population *pop) {
  int i, gene;
  cout << "Generation: " << currGenNo << endl;
  cout << "Individual with best fitness: " << pop->best_individual << endl;
  cout << "Fitness of best individual: " << pop->max_fitness << endl;
  cout << "Mean fitness in generation: " << pop->mean_fitness << endl;

  for (i=0; i<popsize; i++) {
    cout << pop->ind[i].fitness << ": ";
    for (gene=0; gene<lchrom; gene++)
      cout << pop->ind[i].chromosome[gene];
    cout << endl;
  }

  //system("pause");
}


/****************************************************************************
 *		  	Initialization of the population        	    *
 ****************************************************************************/
void initialization(void) {
  int gene, i;

  for (i=0; i<maxgen; i++) {
    log_best_fitness[i] = 0.0;
    log_mean_fitness[i] = 0.0;
    log_mean_diversity[i] = 0.0;
  }

  for (i=0; i < popsize; i++) {
    for (gene=0; gene<lchrom; gene++) 
      popcurr.ind[i].chromosome[gene]= randomBit();

    popcurr.ind[i].fitness = calcFitness(popcurr.ind[i].chromosome);
  }

  if (gaMode == 5) { // ERIGA
    num_elite_immigrants = num_immigrants/2;
    num_rand_immigrants = num_immigrants - num_elite_immigrants;
  }

  if (gaMode == 6) { // HIGA 
    num_elite_immigrants = num_immigrants/3;
    num_dual_immigrants  = num_immigrants/3;
    num_rand_immigrants  = num_immigrants - num_elite_immigrants 
                           - num_dual_immigrants;
  }

  statistics( &popcurr );
  if (bRecordDiversity == true)
    calcPopDiversity( &popcurr );
  //display(&popcurr);
}

/****************************************************************************
 *	  	Execute the GA for one run with index num_exec              *
 ****************************************************************************/
void ga(int num_exec) {
  currGenNo = 0;
  initialization();

  do {
    currGenNo++;  
    if ( (currGenNo%changeSpeed) == 0) 
      changeEnvironment();  // change the problem

    generation();
    statistics( &popnew );
    popcurr = popnew;
    if (bRecordDiversity == true)
      calcPopDiversity( &popcurr );

    //display( &popcurr );
  } while (currGenNo < maxgen);

  saveFitnessResult(num_exec);   // save the fitness result
  if (bRecordDiversity == true)  // save the diversity result
    saveDiversityResult(num_exec);
}

/****************************************************************************
 * Main function: run the GA for totalRuns times        	            *
 ****************************************************************************/
int main(int argc, char *argv[]) {
  int num_exec;

  // read in parameters
  gaMode = atoi(argv[1]);
  functNo = atoi(argv[2]);
  changeMode = atoi(argv[3]);
  changeDegree = atof(argv[4]);
  changeSpeed = atoi(argv[5]);
  ratio_immigrants = atof(argv[6]);
  //pImmigMut = atof(argv[7]);
  //pBaseMut = atof(argv[8]);

  num_immigrants = (int)ceil(ratio_immigrants * popsize);
  if (gaMode == 1) popsize = 100 + (int)(100*ratio_immigrants);
  else popsize = 100;

  if (gaMode<1 || gaMode>6) gaMode = 1;
  if (changeDegree < 0.0) changeDegree = 0.0;
  if (changeDegree > 1.0) bRandRatio = true;
  else bRandRatio = false;

  // Set up dynamic environments
  xorMask = new int[lchrom];
  if (changeMode == 1 || changeMode == 2) {
    if(changeDegree == 0.1) totalBaseStates = 10;
    if(changeDegree == 0.2) totalBaseStates = 5;
    if(changeDegree == 0.5) totalBaseStates = 2;
    if(changeDegree == 1.0) totalBaseStates = 1;
    baseXORMasks = new int*[totalBaseStates];
    for (int i = 0; i < totalBaseStates; i++) 
      baseXORMasks[i] = new int[lchrom];

    constructBaseXORMasks();
  }

  if ( (changeSpeed == 50 || changeSpeed == 100) && 
       changeDegree == 0.2)
    bRecordDiversity = true;

  // Values for global variables
  maxgen = changeSpeed * totalChanges;

  log_mean_diversity = new double[maxgen+1];
  log_mean_fitness = new double[maxgen+1];
  log_best_fitness = new double[maxgen+1];

  if (functNo == 5) { // knapsack problem
    weight = new int[lchrom];
    profit = new int[lchrom];
    initializeKnapsack();
  }

  // Execution
  cout << "*** GA" << gaMode << " is running on function " 
       << functNo << " ***" << endl;

  for (num_exec=0; num_exec < totalRuns; num_exec++) {
    for (int i=0; i<lchrom; i++)
      xorMask[i] = 0;

    srand(num_exec);
    cout << "Execution: " << num_exec << endl;
    ga(num_exec);    	// carry out an execution of GA
  }

  // De-allocation of memory space
  if (changeMode == 1 || changeMode == 2) {
    for (int i = 0; i < totalBaseStates; i++) 
      delete [] baseXORMasks[i];
    delete [] baseXORMasks;
  }

  delete [] log_mean_fitness;
  delete [] log_best_fitness;
  delete [] log_mean_diversity;
  delete [] xorMask;

  if (functNo == 5) {
    delete [] weight;
    delete [] profit;
  }

  cout << "Success!" << endl;
  return 0;
}
